@@ -109,6 +109,10 @@ def generate_coordinates(infile, outfile=None, run_id=None, target_peptides=None
 
             sql_query = f'''
               SELECT
+              FEATURE.RUN_ID as run_id,
+              PEPTIDE.ID as peptide_id,
+              PRECURSOR.ID as precursor_id,
+              TRANSITION.ID as transition_id,
               PEPTIDE.MODIFIED_SEQUENCE AS peptide,
               PRECURSOR.PRECURSOR_MZ AS precursor_mz,
               PRECURSOR.CHARGE AS charge,
@@ -116,6 +120,8 @@ def generate_coordinates(infile, outfile=None, run_id=None, target_peptides=None
               TRANSITION.CHARGE AS product_charge,
               TRANSITION.ANNOTATION AS product_annotation,
               TRANSITION.DETECTING AS product_detecting,
+              FEATURE.EXP_RT - FEATURE.DELTA_RT AS assay_rt,
+              PRECURSOR.LIBRARY_DRIFT_TIME AS assay_im,
               FEATURE.EXP_RT AS rt_apex,
               FEATURE.LEFT_WIDTH AS left_width,
               FEATURE.RIGHT_WIDTH AS right_width,
@@ -149,6 +155,10 @@ def generate_coordinates(infile, outfile=None, run_id=None, target_peptides=None
         else: # if targeted peptides is "None" than select all peptides
             sql_query = f'''
               SELECT
+              FEATURE.RUN_ID as run_id,
+              PEPTIDE.ID as peptide_id,
+              PRECURSOR.ID as precursor_id,
+              TRANSITION.ID as transition_id,
               PEPTIDE.MODIFIED_SEQUENCE AS peptide,
               PRECURSOR.PRECURSOR_MZ AS precursor_mz,
               PRECURSOR.CHARGE AS charge,
@@ -156,6 +166,8 @@ def generate_coordinates(infile, outfile=None, run_id=None, target_peptides=None
               TRANSITION.CHARGE AS product_charge,
               TRANSITION.ANNOTATION AS product_annotation,
               TRANSITION.DETECTING AS product_detecting,
+              FEATURE.EXP_RT - FEATURE.DELTA_RT AS assay_rt,
+              PRECURSOR.LIBRARY_DRIFT_TIME AS assay_im,
               FEATURE.EXP_RT AS rt_apex,
               FEATURE.LEFT_WIDTH AS left_width,
               FEATURE.RIGHT_WIDTH AS right_width,
@@ -163,7 +175,9 @@ def generate_coordinates(infile, outfile=None, run_id=None, target_peptides=None
               {select_feature_ms1_stmt}
               {select_feature_ms2_stmt}
               SCORE_TABLE.QVALUE as qvalue,
-              PRECURSOR.DECOY as decoy
+              PRECURSOR.DECOY as decoy,
+              FEATURE.ID AS feature_id,
+              FEATURE.RUN_ID AS run_id
               FROM PRECURSOR
               INNER JOIN PRECURSOR_PEPTIDE_MAPPING ON PRECURSOR_PEPTIDE_MAPPING.PRECURSOR_ID = PRECURSOR.ID
               INNER JOIN PEPTIDE ON PEPTIDE.ID = PRECURSOR_PEPTIDE_MAPPING.PEPTIDE_ID
@@ -200,6 +214,12 @@ def generate_coordinates(infile, outfile=None, run_id=None, target_peptides=None
         data = data.loc[data.groupby('group_id')['qvalue'].transform(
             'min').eq(data['qvalue'])].reset_index(drop=True)
 
+        # Group and aggregate transition_id into a list
+        agg_transition_id_df = data.groupby(['group_id'])['transition_id'].apply(
+            list).to_frame().reset_index()
+        data = pd.merge(data.drop('transition_id', axis=1),
+                        agg_transition_id_df, on=['group_id'])
+
         # Group and aggregate product m/z into a list
         agg_product_mz_df = data.groupby(['group_id'])['product_mz'].apply(
             list).to_frame().reset_index()
@@ -373,16 +393,36 @@ class data_io():
             spec = self.filtered.getSpectrum(k)
             if spec.getMSLevel() in ms_level:
                 mz, intensity = spec.get_peaks()
+                if (len(mz) == 0 and len(intensity) == 0):
+                    logging.warn(
+                        f"MS{spec.getMSLevel()} spectrum native id {spec.getNativeID()} had no m/z or intensity array, skipping this spectrum")
+                    continue
                 rt = np.full([mz.shape[0]], spec.getRT(), float)
                 im_tmp = spec.getFloatDataArrays()[0]
                 im = im_tmp.get_data()
-                add_df = pd.DataFrame({'native_id': spec.getNativeID(), 'ms_level': spec.getMSLevel(
-                ), 'peptide': spec.getMetaValue('peptide'), 'mz': mz, 'rt': rt, 'im': im, 'int': intensity})
+                precursor = spec.getPrecursors()[0]
+                if self.verbose == 10:
+                    logging.debug(
+                        f"Adding MS{spec.getMSLevel()} spectrum for peptide: {spec.getMetaValue('peptide')} with native id: {spec.getNativeID()}")
+                add_df = pd.DataFrame({'native_id': spec.getNativeID(), 'ms_level': spec.getMSLevel(), 'peptide': spec.getMetaValue(
+                    'peptide'), 'precursor_mz': spec.getMetaValue('precursor_mz'), 'charge': precursor.getCharge(), 'mz': mz, 'rt': rt, 'im': im, 'int': intensity})
+                ## Add additional meta data
+                meta_data = pd.DataFrame()
+                if spec.getMetaValue('rt_apex') is not None:
+                    meta_data['rt_apex'] = pd.Series(spec.getMetaValue('rt_apex'))
+                if spec.getMetaValue('im_apex') is not None:
+                    meta_data['im_apex'] = pd.Series(spec.getMetaValue('im_apex'))
+                if spec.getMetaValue('rt_left_width') is not None:
+                    meta_data['rt_left_width'] = pd.Series(spec.getMetaValue('rt_left_width'))    
+                if spec.getMetaValue('rt_right_width') is not None:
+                    meta_data['rt_right_width'] = pd.Series(spec.getMetaValue('rt_right_width'))
+                if not meta_data.empty:
+                    add_df = add_df.join(meta_data, how='cross')
                 results_df = pd.concat([results_df, add_df])
         return results_df
 
     @method_timer
-    def save_filtered_tsv(self, ms_level=[1], out_file="diapasef_extracted_data.tsv"):
+    def save_filtered_tsv(self, ms_level=[1], out_file="diapasef_extracted_data.tsv", out_type=None):
         """
         Save a tsv file of the targeted filtered spectra in tabular format
 
@@ -391,6 +431,10 @@ class data_io():
           ms_level: (list) list of ms level data to write out to file
           out_file: (str) output file to write data to
         """
+        if out_type is None:
+            out_type = out_file.split(".")[-1]
+        if out_type not in ["tsv", "parquet"]:
+            raise Exception("Output file type must be tsv or parquet")
         results_df = pd.DataFrame()
         for k in tqdm(range(self.filtered.getNrSpectra())):
             spec = self.filtered.getSpectrum(k)
@@ -422,10 +466,15 @@ class data_io():
                 if not meta_data.empty:
                     add_df = add_df.join(meta_data, how='cross')
                 results_df = pd.concat([results_df, add_df])
-        logging.info(
-            f"Saving filtered target spectra data to tabular tsv format: {out_file}")
-        results_df.to_csv(out_file, sep="\t")
-
+        if out_type == "tsv":
+            logging.info(
+                f"Saving filtered target spectra data to tabular tsv format: {out_file}")
+            results_df.to_csv(out_file, sep="\t")
+        elif out_type == "parquet":
+            logging.info(
+                f"Saving filtered target spectra data to parquet format: {out_file}")
+            results_df.to_parquet(out_file, engine='pyarrow')
+            
     def get_target_ms_level_indices(self):
         """
         Extract spectrum indices for a specific mslevel(s).
@@ -500,8 +549,7 @@ class TargeteddiaPASEFExperiment(data_io):
         Return: 
           (tuple) a tuple of the lower and upper bound for target mz
         """
-        # mz_uncertainty = target_mz*(self.mz_tol/(2*1000000))
-        mz_uncertainty = target_mz*(self.mz_tol/(1000000))
+        mz_uncertainty = target_mz * self.mz_tol / 2.0 * 1.0e-6
         target_precursor_mz_upper = target_mz + mz_uncertainty
         target_precursor_mz_lower = target_mz - mz_uncertainty
         return target_precursor_mz_lower, target_precursor_mz_upper
@@ -640,7 +688,7 @@ class TargeteddiaPASEFExperiment(data_io):
                     'peptide', self.peptides[target_peptide_group]['peptide'])
                 spec.setMetaValue(
                     'precursor_mz', self.peptides[target_peptide_group]['precursor_mz'])    
-                [spec.setMetaValue(f'product_mz_{i}', mz) for i, mz in zip(range(0, len(self.peptides[target_peptide_group]['product_mz'])), self.peptides[target_peptide_group]['product_mz'])]
+                [spec.setMetaValue(f'product_mz_{i}', mz) for i, mz in zip(range(0, len(np.unique(self.peptides[target_peptide_group]['product_mz']))), np.unique(self.peptides[target_peptide_group]['product_mz']))]
                 if 'rt_apex' in self.peptides[target_peptide_group].keys():
                     spec.setMetaValue(
                     'rt_apex', self.peptides[target_peptide_group]['rt_apex'])
@@ -723,7 +771,7 @@ class TargeteddiaPASEFExperiment(data_io):
             # Get Coordinates for current peptide
             target_peptide = self.peptides[target_peptide_group]['peptide']
             target_precursor_mz = self.peptides[target_peptide_group]['precursor_mz']
-            target_product_mz = self.peptides[target_peptide_group]['product_mz']
+            target_product_mz = np.unique(self.peptides[target_peptide_group]['product_mz'])
             rt_apex = self.peptides[target_peptide_group]['rt_apex']
             im_apex = self.peptides[target_peptide_group]['im_apex']
 
